// Greg Schaeffer
// ICFP Programming Contest 2001
//
// File: lists.h
//
// Provides struct definitions for linked lists 
// and stacks and functions to work with these
// structures

#ifndef _LISTS_H_
#define _LISTS_H_
#include <stdlib.h>
// struct with whitespace context
struct wsnode
{
   bool tag_b;
   bool tag_em;
   bool tag_i;
   bool tag_s;
   bool tag_tt;
   int tag_u;
   int tag_size;
   int tag_color;
};

// doubly linked list of contexts with a char of text
struct charnode
{
   bool prev_diff;
   wsnode* ws_ptr;
   char text;
   charnode* prev;
   charnode* next;
   bool tag_B;
   bool tag_EM;
   bool tag_I;
   bool tag_S;
   bool tag_TT;
   int tag_U;
   bool tag_0;
   bool tag_1;
   bool tag_2;
   bool tag_3;
   bool tag_4;
   bool tag_5;
   bool tag_6;
   bool tag_7;
   bool tag_8;
   bool tag_9;
   bool tag_r;
   bool tag_g;
   bool tag_b;
   bool tag_c;
   bool tag_m;
   bool tag_y;
   bool tag_k;
   bool tag_w;
   bool tag_PL;
};

// stack of contexts
struct contextnode
{
   bool tag_B;
   bool tag_EM;
   bool tag_I;
   bool tag_S;
   bool tag_TT;
   int tag_U;
   bool tag_0;
   bool tag_1;
   bool tag_2;
   bool tag_3;
   bool tag_4;
   bool tag_5;
   bool tag_6;
   bool tag_7;
   bool tag_8;
   bool tag_9;
   bool tag_r;
   bool tag_g;
   bool tag_b;
   bool tag_c;
   bool tag_m;
   bool tag_y;
   bool tag_k;
   bool tag_w;
   contextnode* next;
};

charnode* char_head = NULL;
charnode* char_tail = NULL;
contextnode* context_head = NULL;
contextnode* context_tail = NULL;

void push_context(contextnode* current)
{
   contextnode* temp = new(contextnode);
   temp -> tag_B = current -> tag_B;
   temp -> tag_EM = current -> tag_EM;
   temp -> tag_I = current -> tag_I;
   temp -> tag_S = current -> tag_S;
   temp -> tag_TT = current -> tag_TT;
   temp -> tag_U = current -> tag_U;
   temp -> tag_0 = current -> tag_0;
   temp -> tag_1 = current -> tag_1;
   temp -> tag_2 = current -> tag_2;
   temp -> tag_3 = current -> tag_3;
   temp -> tag_4 = current -> tag_4;
   temp -> tag_5 = current -> tag_5;
   temp -> tag_6 = current -> tag_6;
   temp -> tag_7 = current -> tag_7;
   temp -> tag_8 = current -> tag_8;
   temp -> tag_9 = current -> tag_9;
   temp -> tag_r = current -> tag_r;
   temp -> tag_g = current -> tag_g;
   temp -> tag_b = current -> tag_b;
   temp -> tag_c = current -> tag_c;
   temp -> tag_m = current -> tag_m;
   temp -> tag_y = current -> tag_y;
   temp -> tag_k = current -> tag_k;
   temp -> tag_w = current -> tag_w;
   temp -> next = NULL;

   if (context_head == NULL)
   {
      context_head = temp;
      context_tail = temp;
   }
   else
   {
      context_tail -> next = temp;
      context_tail = temp;
   }
}

void pop_context(contextnode* current)
{
   contextnode* temp = context_head;
   current -> tag_B = context_tail -> tag_B;
   current -> tag_EM = context_tail -> tag_EM;
   current -> tag_I = context_tail -> tag_I;
   current -> tag_S = context_tail -> tag_S;
   current -> tag_TT = context_tail -> tag_TT;
   current -> tag_U = context_tail -> tag_U;
   current -> tag_0 = context_tail -> tag_0;
   current -> tag_1 = context_tail -> tag_1;
   current -> tag_2 = context_tail -> tag_2;
   current -> tag_3 = context_tail -> tag_3;
   current -> tag_4 = context_tail -> tag_4;
   current -> tag_5 = context_tail -> tag_5;
   current -> tag_6 = context_tail -> tag_6;
   current -> tag_7 = context_tail -> tag_7;
   current -> tag_8 = context_tail -> tag_8;
   current -> tag_9 = context_tail -> tag_9;
   current -> tag_r = context_tail -> tag_r;
   current -> tag_g = context_tail -> tag_g;   
   current -> tag_b = context_tail -> tag_b;
   current -> tag_c = context_tail -> tag_c;
   current -> tag_m = context_tail -> tag_m;
   current -> tag_y = context_tail -> tag_y;
   current -> tag_k = context_tail -> tag_k;
   current -> tag_w = context_tail -> tag_w;

   if (context_head==context_tail)
   {
      delete(temp);
      context_head = NULL;
      context_tail = NULL;
   }
   else
   {
      while (temp->next != context_tail) temp=temp->next;
      context_tail = temp;
      temp = temp -> next;
      context_tail -> next = NULL;
      delete(temp);
   }
}

void init_charnode(contextnode* current)
{
   charnode* temp = new(charnode);
   temp -> tag_B = current -> tag_B;
   temp -> tag_EM = current -> tag_EM;
   temp -> tag_I = current -> tag_I;
   temp -> tag_S = current -> tag_S;
   temp -> tag_TT = current -> tag_TT;
   temp -> tag_U = current -> tag_U;
   temp -> tag_0 = current -> tag_0;
   temp -> tag_1 = current -> tag_1;
   temp -> tag_2 = current -> tag_2;
   temp -> tag_3 = current -> tag_3;
   temp -> tag_4 = current -> tag_4;
   temp -> tag_5 = current -> tag_5;
   temp -> tag_6 = current -> tag_6;
   temp -> tag_7 = current -> tag_7;
   temp -> tag_8 = current -> tag_8;
   temp -> tag_9 = current -> tag_9;
   temp -> tag_r = current -> tag_r;
   temp -> tag_g = current -> tag_g;
   temp -> tag_b = current -> tag_b;
   temp -> tag_c = current -> tag_c;
   temp -> tag_m = current -> tag_m;
   temp -> tag_y = current -> tag_y;
   temp -> tag_k = current -> tag_k;
   temp -> tag_w = current -> tag_w;
   temp -> next = NULL;
   temp -> prev = NULL;
   temp -> text = '>';
   temp -> ws_ptr = NULL;
   temp -> prev_diff = 0;

   char_head = temp;
   char_tail = temp;
}

bool diffchars(charnode* left, charnode* right)
{
   return (left->tag_B != right->tag_B ||
           left->tag_EM != right->tag_EM ||
           left->tag_I != right->tag_I ||
           left->tag_S != right->tag_S ||
           left->tag_TT != right->tag_TT ||
           left->tag_U != right->tag_U ||
           left->tag_0 != right->tag_0 ||
           left->tag_1 != right->tag_1 ||
           left->tag_2 != right->tag_2 ||
           left->tag_3 != right->tag_3 ||
           left->tag_4 != right->tag_4 ||
           left->tag_5 != right->tag_5 ||
           left->tag_6 != right->tag_6 ||
           left->tag_7 != right->tag_7 ||
           left->tag_8 != right->tag_8 ||
           left->tag_9 != right->tag_9 ||
           left->tag_r != right->tag_r ||
           left->tag_g != right->tag_g ||
           left->tag_b != right->tag_b ||
           left->tag_c != right->tag_c ||
           left->tag_m != right->tag_m ||
           left->tag_y != right->tag_y ||
           left->tag_k != right->tag_k ||
           left->tag_w != right->tag_w);
}

bool diffws(wsnode* left, wsnode* right)
{
   if (left==NULL || right==NULL) return 1;

   return (left->tag_b != right->tag_b ||
           left->tag_em != right->tag_em ||
           left->tag_i != right->tag_i ||
           left->tag_s != right->tag_s ||
           left->tag_tt != right->tag_tt ||
           left->tag_size != right->tag_size ||
           left->tag_color != right->tag_color);
}

void push_charnode(contextnode* current, char c)
{
   wsnode* ws = NULL;

   if (c==9 || c==32 || c==13 || c==10)
   {
      ws = new(wsnode);
      ws -> tag_b = 0;
      ws -> tag_em = 0;
      ws -> tag_i = 0;
      ws -> tag_s = 0;
      ws -> tag_size = 10;
      ws -> tag_color = 10;
      if (current->tag_0) ws->tag_size=0;
      if (current->tag_1) ws->tag_size=1;
      if (current->tag_2) ws->tag_size=2;
      if (current->tag_3) ws->tag_size=3;
      if (current->tag_4) ws->tag_size=4;
      if (current->tag_5) ws->tag_size=5;
      if (current->tag_6) ws->tag_size=6;
      if (current->tag_7) ws->tag_size=7;
      if (current->tag_8) ws->tag_size=8;
      if (current->tag_9) ws->tag_size=9;
      ws -> tag_tt = current->tag_TT;
      ws -> tag_u = current->tag_U;

      if (ws->tag_u == 0)
         ws->tag_color = 7;
      else
      {
         if (current->tag_r) ws->tag_color=0;
         if (current->tag_g) ws->tag_color=1;
         if (current->tag_b) ws->tag_color=2;
         if (current->tag_c) ws->tag_color=3;
         if (current->tag_m) ws->tag_color=4;
         if (current->tag_y) ws->tag_color=5;
         if (current->tag_k) ws->tag_color=6;
         if (current->tag_w) ws->tag_color=7;
      }

      if (!ws->tag_tt) c = 32;

      if (!ws->tag_tt && !diffws(ws, char_tail->ws_ptr))
      {
         delete(ws);
         return;
      }
   }

   charnode* temp = new(charnode);
   temp -> tag_B = current -> tag_B;
   temp -> tag_EM = current -> tag_EM;
   temp -> tag_I = current -> tag_I;
   temp -> tag_S = current -> tag_S;
   temp -> tag_TT = current -> tag_TT;
   temp -> tag_U = current -> tag_U;
   temp -> tag_0 = current -> tag_0;
   temp -> tag_1 = current -> tag_1;
   temp -> tag_2 = current -> tag_2;
   temp -> tag_3 = current -> tag_3;
   temp -> tag_4 = current -> tag_4;
   temp -> tag_5 = current -> tag_5;
   temp -> tag_6 = current -> tag_6;
   temp -> tag_7 = current -> tag_7;
   temp -> tag_8 = current -> tag_8;
   temp -> tag_9 = current -> tag_9;
   temp -> tag_r = current -> tag_r;
   temp -> tag_g = current -> tag_g;
   temp -> tag_b = current -> tag_b;
   temp -> tag_c = current -> tag_c;
   temp -> tag_m = current -> tag_m;
   temp -> tag_y = current -> tag_y;
   temp -> tag_k = current -> tag_k;
   temp -> tag_w = current -> tag_w;
   temp -> next = NULL;
   temp -> prev = char_tail;
   char_tail -> next = temp;
   char_tail = temp;
   temp -> text = c;
   temp -> prev_diff = diffchars(temp->prev, temp);
   temp -> ws_ptr = ws;
}
#endif