// Greg Schaeffer
// ICFP 2001 Programming Contest
//
// File: terplist.h
// 
// Provide data structures for the "interpreter"
// for the conversion back to SML/NG.

#ifndef _TERPLIST_H_
#define _TERPLIST_H_

#include "lists.h"
#include <stdlib.h>

struct tagnode
{
   char tag[2];
   tagnode* next;
};

struct textnode
{
   char text;
   textnode* next;
};

struct terpnode
{
   int tag_B;
   int tag_EM;
   int tag_I;
   int tag_S;
   int tag_TT;
   int tag_U;
   int tag_0;
   int tag_1;
   int tag_2;
   int tag_3;
   int tag_4;
   int tag_5;
   int tag_6;
   int tag_7;
   int tag_8;
   int tag_9;
   int tag_r;
   int tag_g;
   int tag_b;
   int tag_c;
   int tag_m;
   int tag_y;
   int tag_k;
   int tag_w;
   int tag_PL;
   terpnode* next;
   terpnode* prev;
   tagnode* head_taga;
   tagnode* tail_taga;
   tagnode* head_tagb;
   tagnode* tail_tagb;
   textnode* head_text;
   textnode* tail_text;
};

terpnode* head_terp=NULL;
terpnode* tail_terp=NULL;

void push_terp(terpnode* x)
{
   terpnode* temp = new(terpnode);
   temp->tag_B = x->tag_B;
   temp->tag_I = x->tag_I;
   temp->tag_EM = x->tag_EM;
   temp->tag_S = x->tag_S;
   temp->tag_TT = x->tag_TT;
   temp->tag_U = x->tag_U;
   temp->tag_0 = x->tag_0;
   temp->tag_1 = x->tag_1;
   temp->tag_2 = x->tag_2;
   temp->tag_3 = x->tag_3;
   temp->tag_4 = x->tag_4;
   temp->tag_5 = x->tag_5;
   temp->tag_6 = x->tag_6;
   temp->tag_7 = x->tag_7;
   temp->tag_8 = x->tag_8;
   temp->tag_9 = x->tag_9;
   temp->tag_r = x->tag_r;
   temp->tag_g = x->tag_g;
   temp->tag_b = x->tag_b;
   temp->tag_c = x->tag_c;
   temp->tag_m = x->tag_m;
   temp->tag_y = x->tag_y;
   temp->tag_k = x->tag_k;
   temp->tag_w = x->tag_w;
   temp->tag_PL = x->tag_PL;
   temp->next = NULL;
   temp->prev = NULL;
   temp->head_taga = NULL;
   temp->tail_taga = NULL;
   temp->head_tagb = NULL;
   temp->tail_tagb = NULL;
   temp->head_text = NULL;
   temp->tail_text = NULL;

   if (head_terp == NULL)
   {
      head_terp = temp;
      tail_terp = temp;
   }
   else
   {
      tail_terp->next = temp;
      temp->prev = tail_terp;
      tail_terp = temp;
   }
   //cout<<"End push_terp"<<endl;
}

void push_tag_back(terpnode* x, char tag[2], bool a)
{
   tagnode* temp = new(tagnode);
   temp->tag[0] = tag[0];
   temp->tag[1] = tag[1];
   temp->next = NULL;

   if (a)
   {
      if (x->head_taga == NULL)
      {
         x->head_taga = temp;
         x->tail_taga = temp;
      }
      else
      {
         x->tail_taga->next = temp;
         x->tail_taga = temp;
      }
   }
   else
   {
      if (x->head_tagb == NULL)
      {
         x->head_tagb = temp;
         x->tail_tagb = temp;
      }
      else
      {
         x->tail_tagb->next = temp;
         x->tail_tagb = temp;
      }
   }
   //cout<<"End push_tag_back "<<tag<<endl;
}

void push_tag_front(terpnode* x, char tag[2], bool a)
{
   tagnode* temp = new(tagnode);
   temp->tag[0] = tag[0];
   temp->tag[1] = tag[1];
   
   if (a)
   {
      if (x->head_taga == NULL)
      {
         x->head_taga = temp;
         x->tail_taga = temp;
         temp->next = NULL;
      }
      else
      {
         temp->next = x->head_taga;
         x->head_taga = temp;
      }
   }
   else
   {
      if (x->head_tagb == NULL)
      {
         x->head_tagb = temp;
         x->tail_tagb = temp;
         temp->next = NULL;
      }
      else
      {
         temp->next = x->head_tagb;
         x->head_tagb = temp;
      }
   }
   //cout<<"End push_tag_front "<<tag<<endl;
}

void push_text(terpnode* x, char text)
{
   textnode* temp = new(textnode);
   temp->text = text;
   temp->next = NULL;

   if (x->head_text == NULL)
   {
      x->head_text = temp;
      x->tail_text = temp;
   }
   else
   {
      x->tail_text->next = temp;
      x->tail_text = temp;
   }
   //cout<<"End push_text "<<text<<endl;
}

void kill_text(textnode* x)
{
   textnode* temp = x;

   while(x != NULL)
   {
      x=x->next;
      temp->next = NULL;
      delete(temp);
      temp=x;
   }
   //cout<<"End kill_text"<<endl;
}

void kill_tag(tagnode* x)
{
   tagnode* temp = x;

   while(x != NULL)
   {
      x=x->next;
      temp->next = NULL;
      delete(temp);
      temp=x;
   }
   //cout<<"End kill_tag"<<endl;
}

void kill_terp()
{
   terpnode* temp = head_terp;

   while(head_terp != NULL)
   {
      head_terp = head_terp->next;
      temp->prev = NULL;
      temp->next = NULL;
      kill_text(temp->head_text);
      temp->head_text = NULL;
      temp->tail_text = NULL;
      kill_tag(temp->head_taga);
      kill_tag(temp->head_tagb);
      temp->head_taga = NULL;
      temp->tail_taga = NULL;
      temp->head_tagb = NULL;
      temp->tail_tagb = NULL;
      delete(temp);
      temp = head_terp;
   }
   //cout<<"End kill_terp"<<endl;
}

#endif